' === SmokerControl.bas =======================================================

' Program monitors and controls the temperature of an electric BBQ smoker.
' It measures the temperature once every second, calculates the average
' every 12 seconds, converts seconds to minutes with one decimal place,
' and sends the time and temperature data to a PC for display and analysis.
' The program also displays real-time temperature on an LCD next to the smoker.

' === Directives ==============================================================
  #com 3                      'specify serial port
  #picaxe 08M2                'specify processor
  #terminal 4800              'open terminal window
  #no_data                    'save time downloading

' === Constants ===============================================================
  symbol     baud = N4800_16  'baud rate for terminal and LCD
  symbol      TxD = C.0       'data transmitted to terminal on pin C.0 
  symbol      LCD = C.1       'data output to LCD  on pin C.1
  symbol     heat = C.2       'output to SSR for heating element on pin C.2
  symbol    therm = C.4       'analog data input from thermocouple on pin C.4
  symbol      RxD = C.5       'data input from terminal on pin C.5 
      
' HD44780 LCD Commands
  symbol      clr =   1       'clear screen, and cursor home
  symbol       DS =   7       'location of LCD custom character for degree sign
  symbol       L1 = 128       'move cursor to line 1, position 1
  symbol       L2 = 192       'move cursor to line 2, position 1
  symbol       TP = 196       'move cursor to position to update temperature
    
' === Variables ===============================================================
' Program needs more than the 28 GP variables that are available on the 08M2.   
' To solve this problem, the first five variables are given "generic" names
' so they can be re-used for 2 different purposes that don't require long-term
' data storage: the processing of user input at the beginning of the program,
' and the bintoascii conversions in the logData Subroutine.
  Symbol     var0 = b0
  Symbol     var1 = b1
  Symbol     var2 = b2
  Symbol     var3 = b3
  Symbol     var4 = b4
  
  symbol    index = b5        'used in for/next loop
  symbol  samples = b6        'number of tempF readings averaged in main loop
  symbol      now = w4        'the current second
  symbol   minute = w5        'used to log current time in minutes 
  symbol nextTime = w6        'used to implement 2-second main loop time 
  symbol    tempC = w7        'Centigrade real-time temperature
  symbol    tempF = w8        'Fahrenheit real-time temperature
  symbol  targetF = w9        'Fahrenheit target temperature
  symbol  minHeat = w10       'minimum heat time needed to maintain target temp
  symbol    error = w11       'the difference between targetF & tempF
  
' tempSum is used to average multiple temp readings in the main program
  symbol  tempSum = w12
  
' tempTot is used to average multiple temp readings in the getTempF subroutine
  symbol  tempTot = w13
  
' === Begin Main Program ======================================================
adcconfig %011                             'Vref- = gnd, Vref+ = FVR
setfreq m16                                'so time variable works correctly
pullup %01000                              'enable internal pullup on pin C.3
wait 4                                     'allow time for LCD to initialize

serout LCD,baud,(L1,"Ready to",L2," Smoke! ")
serout TxD,baud,("Ready to Smoke!",cr,lf,cr,lf)

disconnect                                 'so we can use C.0 for serin command

' Get target temperature from the user and check input for correctness.
' Loop back to this point until the user gets it right!  :-) 
tryAgain:
var2=255                                   'prepare for input error checking
var1=255
var0=255

serout TxD,baud,("Please enter a three-digit target temperature",cr,lf)
serout TxD,baud,("between 100 and 250 degrees, inclusive.",cr,lf,cr,lf)
serin  RxD,baud,var2                       'blocking
serin  [400],RxD,baud,var1,var0            'non-blocking: 100mS timeout

var2 = var2 - 48                           'convert ASCII to digit
var1 = var1 - 48                           'convert ASCII to digit
var0 = var0 - 48                           'convert ASCII to digit

if var2 > 2 OR var1 > 9 OR var0 > 9 then
  serout TxD,baud,("Invalid Input: Try Again!",cr,lf,cr,lf)
  goto tryAgain
endif
 
var2 = 100 * var2                          'convert digit to decimal value
var1 =  10 * var1                          'convert digit to decimal value
targetF = var2 + var1 + var0               'target value in whole degrees

if targetF < 100 or targetF > 250 then
  serout TxD,baud,("Invalid Input: Try Again!",cr,lf,cr,lf)
  goto tryAgain
endif

reconnect                                  'we no longer need serin command

' Display target temperature in the terminal window.
serout TxD,baud,("Target temperature is set at ")
serout TxD,baud,(#targetF," degrees F.",cr,lf,cr,lf)

' Also display target temperature on the LCD.  
serout LCD,baud,(clr,"Set ",#targetF,DS)

' *** Note ********************************************************************
' The program statement following this note is set correctly for the first test
' in the PICAXE Primer column. As explained in the text, the value of minHeat
' will be changed for each subsequent test.
'
' When all the necessary tests have been completed, and an effective control
' algorithm has been developed, the following program statement can be replaced
' by the algorithm. That way, the program will function correctly at any chosen
' target temperature between 100 and 250 degrees Fahrenheit.
' *****************************************************************************

' Compute minimum heat time for target temperature.
minHeat = 1125

' Send targetF and minHeat values to terminal.
serout TxD,baud,("targetF = ",#targetF,cr,lf)
serout TxD,baud,("minHeat = ",#minHeat,cr,lf,cr,lf)

' Send column headers to terminal.
serout TxD,baud,(" Time, Temp",cr,lf)

targetF = 10 * targetF                     'Convert target temp to tenths

' get the first temperature reading
gosub getTempF
time = 0
gosub logData
tempTot = 0
samples = 0
time = 1

' === Main Program Loop =======================================================
do
  nextTime = time + 1                      'for a 1-second loop time  
  gosub getTempF  
  tempTot = tempTot + tempF
  inc samples
  
  if tempF < targetF then
     error = targetF - tempF
  else
    error = 0
  endif

  gosub controlTemp

  now = samples // 12
  
  if now = 0 then
    tempF = tempTot / samples
    gosub logData
    tempTot = 0
    samples = 0
  endif

  gosub displayTempF                       'must execute after gosub logData
  tarry: if time < nextTime then tarry     'wait for remainder of two seconds
loop

' === End Main Program - Subroutines Follow ===================================

' === controlTemp Subroutine ==================================================
' Calculate length of time heat is on during each iteration of the main loop.
controlTemp:
  select case error
    case > 475                             'error > 47.5 degrees F
      high heat                            'heat on for 100% of loop time
    case > 0                               '10 degrees F > error > 0 degrees F
      high heat
      error = 8 * error MIN minHeat        'starts at 3800 (almost 100%),
      pause error                          'and decreases to minHeat
      low heat                             'as error decreases to 0
    case = 0                               'in this case, we're at target temp,
      low heat                             'so turn off heat
  endselect
return

' === displayTempF Subroutine =================================================
' Display the current smoker temperature (in degrees F) on line 2 of the LCD.
displayTempF:
  tempF = tempF + 5                        'round tempF to nearest degree
  tempF = tempF / 10
  
  serout LCD,baud,(L2,"Now ")
  
  if tempF < 100 then
    serout LCD,baud,(" ")
  endif
  
  serout LCD,baud,(#tempF,DS)
return

' === getTempF Subroutine =====================================================
' Get the current temperature in tenths of a degree Fahrenheit.
getTempF:
  tempSum = 0
  
  for index = 1 to 10
    fvrsetup FVR4096                       'FVR = 4096mV
    readadc10 therm, tempC                 'get raw ADC value
    tempC = 4 * tempC                      'convert ADCval to mV
    tempC = 2 * tempC - 2500               'convert mV to tempC in tenths
    tempSum = tempSum + tempC
  next index

  tempC = tempSum / 10                     'get tempC in tenths
  tempF = 9 * tempC / 5 + 320              'convert to tempF in tenths
return

' === logData Subroutine ======================================================
' Send the data to terminal program in csv format.
logData:
  minute = time / 6                        'current minute in tenths

' Separate the digits in the minute variable, & send each digit to the terminal  
  bintoascii minute,var4,var3,var2,var1,var0
  serout TxD,baud,(var3,var2,var1,".",var0,",")
  
' Separate the digits in the tempF variable, & send each digit to the terminal    
  bintoascii tempF,var4,var3,var2,var1,var0
  serout TxD,baud,(var3,var2,var1,".",var0,cr,lf)   
return